from scurrypy import (
    Addon,
    EmojiModel,
    Button, ButtonStyles,
    ActionRowChild, ActionRowPart,
    SelectOption,
    TextInput, TextInputStyles,
    DefaultValue
)

class ComponentBuilder(Addon):

    @staticmethod
    def _basic_button(
        style: int, 
        custom_id: str, 
        label: str = None, 
        emoji: str | EmojiModel = None, 
        disabled: bool = False
    ):
        if emoji:
            if isinstance(emoji, str):
                emoji = EmojiModel(name=emoji)
            elif not isinstance(emoji, EmojiModel):
                raise TypeError(f"EasyBot.primary expects type str or EmojiModel, got {type(emoji).__name__}")
        
        return Button(
            style=style,
            custom_id=custom_id,
            label=label,
            emoji=emoji ,
            disabled=disabled
        )
    
    @staticmethod
    def primary(
        custom_id: str, 
        label: str = None, 
        emoji: str | EmojiModel = None, 
        disabled: bool = False
    ):
        """Builds a primary button

        Args:
            custom_id (str): unique button identifier
            label (str, optional): user-facing label
            emoji (str | EmojiModel, optional): emoji icon as str or EmojiModel if custom
            disabled (bool, optional): Whether the button should be disabled. Defaults to False.

        Returns:
            (Button): the button object
        """
        return ComponentBuilder._basic_button(ButtonStyles.PRIMARY, custom_id, label, emoji, disabled)
    
    @staticmethod
    def secondary(
        custom_id: str, 
        label: str = None, 
        emoji: str | EmojiModel = None, 
        disabled: bool = False
    ):
        """Builds a secondary button

        Args:
            custom_id (str): unique button identifier
            label (str, optional): user-facing label
            emoji (str | EmojiModel, optional): emoji icon as str or EmojiModel if custom
            disabled (bool, optional): Whether the button should be disabled. Defaults to False.

        Returns:
            (Button): the button object
        """
        return ComponentBuilder._basic_button(ButtonStyles.SECONDARY, custom_id, label, emoji, disabled)
    
    @staticmethod
    def success(
        custom_id: str, 
        label: str = None, 
        emoji: str | EmojiModel = None, 
        disabled: bool = False
    ):
        """Builds a success button

        Args:
            custom_id (str): unique button identifier
            label (str, optional): user-facing label
            emoji (str | EmojiModel, optional): emoji icon as str or EmojiModel if custom
            disabled (bool, optional): Whether the button should be disabled. Defaults to False.

        Returns:
            (Button): the button object
        """
        return ComponentBuilder._basic_button(ButtonStyles.SUCCESS, custom_id, label, emoji, disabled)
    
    @staticmethod
    def danger(
        custom_id: str, 
        label: str = None, 
        emoji: str | EmojiModel = None, 
        disabled: bool = False
    ):
        """Builds a danger button

        Args:
            custom_id (str): unique button identifier
            label (str, optional): user-facing label
            emoji (str | EmojiModel, optional): emoji icon as str or EmojiModel if custom
            disabled (bool, optional): Whether the button should be disabled. Defaults to False.

        Returns:
            (Button): the button object
        """
        return ComponentBuilder._basic_button(ButtonStyles.DANGER, custom_id, label, emoji, disabled)
    
    @staticmethod
    def link(
        url: str, 
        label: str = None, 
        emoji: str | EmojiModel = None, 
        disabled: bool = False
    ):
        """Builds a link button

        Args:
            url (str): button URL to open
            label (str, optional): user-facing label
            emoji (str | EmojiModel, optional): emoji icon as str or EmojiModel if custom
            disabled (bool, optional): Whether the button should be disabled. Defaults to False.

        Returns:
            (Button): the button object
        """
        btn = ComponentBuilder._basic_button(ButtonStyles.LINK, label, emoji, disabled)
        btn.url = url
        return btn
    
    @staticmethod
    def row(components: list[ActionRowChild]):
        """Shorthand for action row.

        Args:
            components (list[ActionRowChild]): the action row objects.

        Returns:
            (ActionRowPart): the action row object
        """
        if not isinstance(components, list):
            components = [components]

        return ActionRowPart(components)
    
    @staticmethod
    def option(
        label: str,
        value: str,
        description: str = None,
        emoji: EmojiModel | str = None,
        default: bool = False
    ):
        """Builds a string menu select option.

        Args:
            label (str): user-facing label
            value (str): unique identifier
            description (str, optional): option descriptor
            emoji (EmojiModel | str, optional): emoji icon as str or EmojiModel if custom
            default (bool, optional): Whether this value should be the default if none is selected. Defaults to False.

        Raises:
            (TypeError): invalid `emoji` type

        Returns:
            (SelectOption): the SelectOption object
        """
        if emoji:
            if isinstance(emoji, str):
                emoji = EmojiModel(name=emoji)
            elif not isinstance(emoji, EmojiModel):
                raise TypeError(f"EasyBot.primary expects type str or EmojiModel, got {type(emoji).__name__}")
        
        return SelectOption(
            label=label,
            value=value,
            description=description,
            emoji=emoji,
            default=default
        )
    
    @staticmethod
    def short_text(
        custom_id: str,
        required: bool = True,
        placeholder: str | None = None,
        min_length: int | None = None,
        max_length: int | None = None
    ):
        """Builds a TextInput with `TextInputStyles.SHORT` style.

        Args:
            custom_id (str): unique identifier
            required (bool, optional): Whether this field is required. Defaults to True.
            placeholder (str | None, optional): default text if nothing is entered
            min_length (int | None, optional): minimum input length
            max_length (int | None, optional): maximum input length

        Returns:
            (TextInput): the TextInput object
        """
        return TextInput(
            style=TextInputStyles.SHORT,
            custom_id=custom_id,
            required=required,
            placeholder=placeholder,
            min_length=min_length,
            max_length=max_length
        )

    @staticmethod
    def long_text(
        custom_id: str,
        required: bool = True,
        placeholder: str | None = None,
        min_length: int | None = None,
        max_length: int | None = None
    ):
        """Builds a TextInput with `TextInputStyles.PARAGRAPH` style.

        Args:
            custom_id (str): unique identifier
            required (bool, optional): Whether this field is required. Defaults to True.
            placeholder (str | None, optional): default text if nothing is entered
            min_length (int | None, optional): minimum input length
            max_length (int | None, optional): maximum input length

        Returns:
            (TextInput): the TextInput object
        """
        return TextInput(
            style=TextInputStyles.PARAGRAPH,
            custom_id=custom_id,
            required=required,
            placeholder=placeholder,
            min_length=min_length,
            max_length=max_length
        )

    @staticmethod
    def role_value(id: int):
        """Builds a default role value.

        Args:
            id (int): role ID

        Returns:
            (DefaultValue): the DefaultValue object
        """
        return DefaultValue(
            id=id,
            type='role'
        )

    @staticmethod
    def user_value(id: int):
        """Builds a default user value.

        Args:
            id (int): user ID

        Returns:
            (DefaultValue): the DefaultValue object
        """
        return DefaultValue(
            id=id,
            type='user'
        )
    
    @staticmethod
    def channel_value(id: int):
        """Builds a default channel value.

        Args:
            id (int): channel ID

        Returns:
            (DefaultValue): the DefaultValue object
        """
        return DefaultValue(
            id=id,
            type='channel'
        )



from typing import TypedDict, Unpack

from scurrypy.client import Client
from scurrypy.core import *

from scurrypy.parts.command import CommandOption

from .addons.cache import *

class CacheConfig(TypedDict, total=False):
    """Types of objects to cache."""
    roles: bool
    channels: bool
    bot_emojis: bool
    guild_emojis: bool

class ScurryKit(Client):
    """Prepackaged interaction, prefix, and event convenience for most of your needs!"""

    def __init__(self, *, 
        token: str, 
        application_id: int, 
        intents = Intents.DEFAULT, 
        prefix = None, 
        sync_commands = True, 
        debug_mode = False,
        **cache_config: Unpack[CacheConfig]
    ):
        """
        Args:
            token (str): the bot's token
            application_id (int): the bot's user ID
            intents (int, optional): gateway intents. Defaults to `Intents.DEFAULT`.
            prefix (str, optional): prefix if using prefix commands
            sync_commands (bool, optional): Whether commands should be synced with changes. Defaults to `True`.
            debug_mode (bool, optional): Whether error trace should be printed. Defaults to `False`.
            **cache_config (Unpack[CacheConfig], optional): what objects should be cached
        """

        # THEN init client
        super().__init__(
            token=token, 
            application_id=application_id, 
            intents=intents,
            logger=Logger(debug_mode=debug_mode)
        )

        # init addons
        self.channel_cache = GuildChannelCacheAddon(self) if cache_config.get('channels') else None
        self.role_cache = RoleCacheAddon(self) if cache_config.get('roles') else None
        self.bot_emoji_cache = BotEmojisCacheAddon(self) if cache_config.get('bot_emojis') else None
        self.guild_emoji_cache = GuildEmojiCacheAddon(self) if cache_config.get('guild_emojis') else None

        from .addons.prefix import PrefixAddon
        self.prefixes = PrefixAddon(self, prefix) if prefix else None

        from .addons.interaction import InteractionAddon
        self.commands = InteractionAddon(self, sync_commands)

        from .addons.events import EventsAddon
        self.bot_events = EventsAddon(self)

        self._startup_hooks = []
        self._shutdown_hooks = []

        self.add_startup_hook(self.run_startup_hooks)
        self.add_shutdown_hook(self.run_shutdown_hooks)
        
    async def get_channel(self, channel_id: int):
        """Fetch a guild channel. If not found, request and store it.

        Args:
            channel_id (int): ID of channel

        Returns:
            (ChannelModel | None): hydrated channel object or None if fetch failed
        """
        channel = self.channel_cache.get_channel(channel_id)
        if channel:
            return channel

        try:
            channel = await self.channel(channel_id).fetch()
        except DiscordError:
            return None

        self.channel_cache.put(channel)
        return channel
    
    async def get_role(self, guild_id: int, role_id: int):
        """Fetch a guild role. If not found, request and store it.

        Args:
            guild_id (int): guild ID of role
            role_id (int): role ID of guild

        Returns:
            (RoleModel | None): hydrated role object or None if fetch failed
        """
        role = self.role_cache.get_role(role_id)
        if role:
            return role
        
        try:
            role = await self.guild(guild_id).fetch_guild_role(role_id)
        except DiscordError:
            return None
        
        self.role_cache.put(guild_id, role)
        return role
    
    def get_guild_emoji(self, emoji_id: int):
        """Get a guild emoji by ID.

        Args:
            emoji_id (int): emoji ID

        Returns:
            (EmojiModel | None): the queried emoji object if in cache else None
        """
        return self.guild_emoji_cache.get_emoji(emoji_id)
    
    def get_bot_emoji(self, emoji_name: str):
        """Get a bot emoji by ID.

        Args:
            emoji_name (str): name of emoji

        Returns:
            (EmojiModel | None): the queried emoji object if in cache else None
        """
        return self.bot_emoji_cache.get_emoji(emoji_name)

    async def run_startup_hooks(self):
        """Wrapper for running user defined start hooks."""
        for hook in self._startup_hooks:
            await hook(self)

    async def run_shutdown_hooks(self):
        """Wrapper for running user defined end hooks."""
        for hook in self._shutdown_hooks:
            await hook(self)

    def start_hook(self, func):
        """Decorator to register a startup hook."""
        self._startup_hooks.append(func)
        return func
    
    def end_hook(self, func):
        """Decorator to register a shutdown hook."""
        self._shutdown_hooks.append(func)
        return func

    def prefix(self, name: str):
        """Register a prefix command using PrefixAddon.

        Args:
            name (str): name of the command
                !!! warning "Important"
                    Prefix commands are CASE-INSENSITIVE.
        """
        if not self.prefixes:
            raise AttributeError("Prefixes Addon is not set. Consider setting a prefix.")

        return self.prefixes.register(name)
    
    def event(self, event_name: str):
        """Register an event in which to listen using EventsAddon.

        Args:
            event_name (str): event name
        """
        return self.bot_events.event(event_name.upper())
    
    def slash_command(self, name: str, description: str, *, options: list[CommandOption] = None, guild_ids: list[int] = None):
        """Register and route a slash command using InteractionAddon.

        Args:
            name (str): command name
            description (str): command description
            options (list[CommandOption], optional): list of command options
            guild_ids (list[int], optional): list of guild IDs for guild commands or omit for global
        """
        return self.commands.slash_command(name, description, options=options, guild_ids=guild_ids)
    
    def user_command(self, name: str, guild_ids: list[int] = None):
        """Register and route a user command using InteractionAddon.

        Args:
            name (str): command name
            guild_ids (list[int], optional): list of guild IDs for guild commands or omit for global
        """
        return self.commands.user_command(name, guild_ids=guild_ids)
    
    def message_command(self, name: str, guild_ids: list[int] = None):
        """Register and route a message command using InteractionAddon.

        Args:
            name (str): command name
            guild_ids (list[int], optional): list of guild IDs for guild commands or omit for global
        """
        return self.commands.message_command(name, guild_ids=guild_ids)
    
    def autocomplete(self, command_name: str, option_name: str):
        """Register and route an autocomplete interaction using InteractionAddon.

        Args:
            command_name (str): name of command to autocomplete
            option_name (str): name of option to autocomplete
        """
        return self.commands.autocomplete(command_name, option_name)

    def button(self, custom_id: str):
        """Decorator to route button interactions using InteractionAddon.

        Args:
            custom_id (str): custom ID of button
        """
        return self.commands.button(custom_id)
    
    def select(self, custom_id: str):
        """Decorator to route select menu interactions using InteractionAddon.

        Args:
            custom_id (str): custom ID of select menu
        """
        return self.commands.select(custom_id)
    
    def modal(self, custom_id: str):
        """Decorator to route modal interactions using InteractionAddon.

        Args:
            custom_id (str): custom ID of modal
        """
        return self.commands.modal(custom_id)
